/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.nas.facade;

import com.sun.xml.ws.dump.MessageDumpingFeature;
import dk.sundhedsdatastyrelsen.cbs.dgws.DGWSRequestFactory;
import dk.sundhedsdatastyrelsen.cbs.dgws.STSConfig;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpExecutor;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.JobStatus;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.JobStatusNotification;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.ObjectFactory;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.junit.jupiter.api.*;
import org.oasis_open.docs.wsn.b_2.DestroyPullPoint;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;
import org.oasis_open.docs.wsn.b_2.Unsubscribe;
import org.oasis_open.docs.wsn.bw_2.*;
import org.xml.sax.InputSource;

import javax.xml.bind.JAXBException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Holder;
import javax.xml.ws.wsaddressing.W3CEndpointReference;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.List;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.assertEquals;

@CbsTest
@Disabled("Calls NAS")
class NasEndToEndIT {
    private final Logger logger = Logger.getLogger(NasEndToEndIT.class);
    public static final String NOTIFICATIONBROKER_SERVICE = "https://test1.ekstern-test.nspop.dk:8443/nas2/notificationbroker/service";
    public static final String PULLPOINTFACTORY_SERVICE = "https://test1.ekstern-test.nspop.dk:8443/nas2/pullpointfactory/service";
    public static final String SUBSCRIPTIONMANAGER_SERVICE = "https://test1.ekstern-test.nspop.dk:8443/nas2/subscriptionmanager/service";
    public static final String JOB_COMPLETED_TOPIC = "http://nspop.dk/GreenCertificate/2021/04/01:JobCompleted";

    private NasPublisherImpl nasPublisher;
    private DGWSHeaderProvider headerProvider;
    private PullPointFactory pullPointFactory;
    private W3CEndpointReference pullPoint;
    private W3CEndpointReference subscribe;
    private NasClientImpl nasClient;
    private final MessageDumpingFeature messageDumpingFeature = new MessageDumpingFeature(true, NasEndToEndIT.class.getName(), Level.INFO.toString(), true);

    @BeforeEach
    void setUp() throws JAXBException {
        messageDumpingFeature.disableMessageLogging();
        STSConfig stsConfig = new STSConfig(
            "http://test1.ekstern-test.nspop.dk:8080/sts/services/NewSecurityTokenService",
            "Sundhedsdatastyrelsen",
            "33257872",
            "Sundhedsdatastyrelsen - COVID-19 Borgerstatus.jks",
            "Test1234",
            "CBS"
        );

        NotificationConsumer soap = new NotificationBrokerService().getSoap();
        setEndpoint((BindingProvider) soap, NOTIFICATIONBROKER_SERVICE);

        nasPublisher = new NasPublisherImpl(soap, JOB_COMPLETED_TOPIC);
        headerProvider = new DefaultDGWSHeaderProvider(new DGWSRequestFactory(stsConfig, new dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientServiceImpl(new HttpExecutor())));
        pullPointFactory = new PullPointFactory();


        NotificationProducer notificationProducer = new SubscriptionManagerService().getSoap();
        setEndpoint((BindingProvider) notificationProducer, SUBSCRIPTIONMANAGER_SERVICE);

        CreatePullPoint createPullPoint = new PullPointFactoryService().getSoap();
        setEndpoint((BindingProvider) createPullPoint, PULLPOINTFACTORY_SERVICE);

        nasClient = new NasClientImpl(headerProvider, notificationProducer, createPullPoint, pullPointFactory);
    }

    @AfterEach
    void tearDown() throws Exception {
        if (subscribe != null) {
            String subscribeAddress = EndpointReferences.extractAddress(subscribe);
            NotificationProducer notificationProducer = new SubscriptionManagerService().getSoap();
            setEndpoint((BindingProvider) notificationProducer, subscribeAddress);
            DGWSHeaders headers = headerProvider.getHeaders();
            notificationProducer.unsubscribe(new Unsubscribe(), headers.getSecurity(), new Holder<>(headers.getMedComHeader()));
            pullPointFactory.getPullPoint(pullPoint).destroyPullPoint(new DestroyPullPoint(), headers.getSecurity(), new Holder<>(headers.getMedComHeader()));
        }
    }

    @Test
    void create_PullPoint() {
        NotificationProducer notificationProducer = new SubscriptionManagerService().getSoap();
        setEndpoint((BindingProvider) notificationProducer, SUBSCRIPTIONMANAGER_SERVICE);
        CreatePullPoint createPullPoint = new PullPointFactoryService().getSoap();
        setEndpoint((BindingProvider) createPullPoint, PULLPOINTFACTORY_SERVICE);
        NasClientImpl nasClient = new NasClientImpl(headerProvider, notificationProducer, createPullPoint, pullPointFactory);
        pullPoint = nasClient.createPullPoint();
        Assertions.assertNotNull(pullPoint);
        nasClient.subscribe(pullPoint, JOB_COMPLETED_TOPIC);
        logger.info(EndpointReferences.extractAddress(pullPoint));
    }

    @Test
    void get_messages() {
        //insert url received from @Test create_PullPoint() to poll NAS for messages since that PullPoint was made
        String pullPointAddress = "http://test1.ekstern-test.nspop.dk:8080/nas2/pullpoint/service/3d2f4450-792e-44cf-bd16-d992c9db280d";

        NotificationProducer notificationProducer = new SubscriptionManagerService().getSoap();
        setEndpoint((BindingProvider) notificationProducer, SUBSCRIPTIONMANAGER_SERVICE);
        CreatePullPoint createPullPoint = new PullPointFactoryService().getSoap();
        setEndpoint((BindingProvider) createPullPoint, PULLPOINTFACTORY_SERVICE);
        NasClientImpl nasClient = new NasClientImpl(headerProvider, notificationProducer, createPullPoint, pullPointFactory);

        final W3CEndpointReference pullPoint = EndpointReferences.newEndpointReference(pullPointAddress);

        List<NotificationMessageHolderType> notifications = nasClient.getMessages(pullPoint, 1000, UUID.randomUUID().toString(), messageDumpingFeature);
        Assertions.assertNotNull(notifications);
        logger.info("Found " + notifications.size() + " NAS notifications");
        printSoapResponses();
    }

    @Test
    void subscribe_and_publish_and_get_messages() {
        NotificationProducer notificationProducer = new SubscriptionManagerService().getSoap();
        setEndpoint((BindingProvider) notificationProducer, SUBSCRIPTIONMANAGER_SERVICE);

        CreatePullPoint createPullPoint = new PullPointFactoryService().getSoap();
        setEndpoint((BindingProvider) createPullPoint, PULLPOINTFACTORY_SERVICE);

        NasClientImpl nasClient = new NasClientImpl(headerProvider, notificationProducer, createPullPoint, pullPointFactory);

        pullPoint = nasClient.createPullPoint();
        subscribe = nasClient.subscribe(pullPoint, JOB_COMPLETED_TOPIC);

        JobStatusNotification jobStatusNotification = createJobStatusNotification();

        nasPublisher.publish(jobStatusNotification, "0506504003");

        List<NotificationMessageHolderType> messages = nasClient.getMessages(pullPoint, 1000, UUID.randomUUID().toString(), messageDumpingFeature);

        Assertions.assertNotNull(messages);

        logger.info("Found " + messages.size() + " messages");
        printSoapResponses();
    }

    @Test
    void test_message_replay() {
        pullPoint = nasClient.createPullPoint();
        subscribe = nasClient.subscribe(pullPoint, JOB_COMPLETED_TOPIC);

        JobStatusNotification jobStatusNotification = createJobStatusNotification();

        nasPublisher.publish(jobStatusNotification, "0506504003");
        nasPublisher.publish(jobStatusNotification, "0506504003");
        nasPublisher.publish(jobStatusNotification, "0506504003");
        nasPublisher.publish(jobStatusNotification, "0506504003");

        String messageId = UUID.randomUUID().toString();

        List<NotificationMessageHolderType> originalMessages = nasClient.getMessages(pullPoint, 1000, messageId);

        assertEquals(4, originalMessages.size());

        // replay request using same message id
        List<NotificationMessageHolderType> replayMessages = nasClient.getMessages(pullPoint, 1000, messageId);

        assertEquals(4, replayMessages.size());

        // use new message id
        List<NotificationMessageHolderType> newMessages = nasClient.getMessages(pullPoint, 1000, UUID.randomUUID().toString());

        assertEquals(0, newMessages.size());
    }

    private String formatXml(String xml) {
        try {
            Transformer serializer = SAXTransformerFactory.newInstance().newTransformer();
            serializer.setOutputProperty(OutputKeys.INDENT, "yes");
            serializer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
            serializer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
            //serializer.setOutputProperty("{http://xml.customer.org/xslt}indent-amount", "2");
            Source xmlSource = new SAXSource(new InputSource(new ByteArrayInputStream(xml.getBytes())));
            StreamResult res = new StreamResult(new ByteArrayOutputStream());
            serializer.transform(xmlSource, res);
            return res.getOutputStream().toString();
        } catch (Exception e) {
            return xml;
        }
    }

    private JobStatusNotification createJobStatusNotification() {
        ObjectFactory objectFactory = new ObjectFactory();
        JobStatusNotification jobStatusNotification = objectFactory.createJobStatusNotification();
        jobStatusNotification.setJobId(UUID.randomUUID().toString());
        jobStatusNotification.setJobStatus(JobStatus.OK);
        return jobStatusNotification;
    }

    private void printSoapResponses() {
        String nasMessage = messageDumpingFeature.nextMessage();
        while (nasMessage != null) {
            if (nasMessage.startsWith("Response message received")) {
                String nasMessageXml = StringUtils.substringAfter(nasMessage, "<?xml version='1.0' encoding='UTF-8'?>");
                logger.info("Got Message: \n" + formatXml(nasMessageXml));
            }
            nasMessage = messageDumpingFeature.nextMessage();
        }
    }

    private void setEndpoint(BindingProvider bindingProvider, String url) {
        bindingProvider.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, url);
    }
}
